from .http import HTTPError


async def ensure_network(docker, name):
    try:
        await docker.create_network(data={
            'Name': name,
            'Driver': 'bridge',
            'CheckDuplicate': True,
        })
    except HTTPError as err:
        if err.reason == 'Conflict':
            return
        raise



import sys

from operator import attrgetter
from collections import defaultdict
from collections import Counter, namedtuple

from ..run import run
from ..types import DockerImage, Mode, LocalPath
from ..utils import format_size
from ..images import pull, push, docker_image
from ..images import image_versions
from ..status import Status
from ..console import pretty, config_tty
from ..resolve import resolve

from .._requires import click

from .common import ExtGroup, AsyncCommand


@click.command('build', help='Build image', cls=AsyncCommand)
@click.argument('name')
@click.pass_obj
async def image_build(env, name):
    image = env.images.get(name)
    with Status() as status:
        failed = await resolve(
            env.docker,
            env.images,
            env.services,
            image,
            status=status,
            pull=True,
            build=True,
        )
    if failed:
        click.echo('Failed to build image {}'.format(name))
        sys.exit(1)


def _get_image(images_map, name):
    try:
        image = images_map.get(name)
    except KeyError:
        return DockerImage(name)
    else:
        return docker_image(images_map, image.name)


@click.command('info', help='Show image info', cls=AsyncCommand)
@click.argument('name')
@click.option('--repo-tag', is_flag=True)
@click.pass_obj
async def image_info(env, name, repo_tag):
    try:
        image = env.images.get(name)
    except KeyError:
        click.echo('Unknown image name: {}'.format(name))
        sys.exit(1)
    if repo_tag:
        version, = image_versions(env.images, [image])
        click.echo('{}:{}'.format(image.repository, version))
    else:
        click.echo('Nothing')
        sys.exit(1)


@click.command('pull', help='Pull image version', cls=AsyncCommand)
@click.argument('name')
@click.pass_obj
async def image_pull(env, name):
    image = _get_image(env.images, name)
    with Status() as status:
        success = await pull(env.docker, image, status=status)
    if not success:
        click.echo('Unable to pull image {}'.format(image.name))
        sys.exit(1)


@click.command('push', help='Push image version', cls=AsyncCommand)
@click.argument('name')
@click.pass_obj
async def image_push(env, name):
    image = _get_image(env.images, name)
    with Status() as status:
        success = await push(env.docker, image, status=status)
    if not success:
        click.echo('Unable to push image {}'.format(image.name))
        sys.exit(1)


@click.command('run', help='Run command in container', cls=AsyncCommand)
@click.argument('name')
@click.argument('args', nargs=-1, required=True)
@click.pass_obj
async def image_run(env, name, args):
    image = _get_image(env.images, name)
    volumes = [LocalPath('.', '.', Mode.RW)]

    with config_tty() as tty:
        exit_code = await run(env.docker, tty, image, args,
                              volumes=volumes, work_dir='.')
        sys.exit(exit_code)


_Tag = namedtuple('_Tag', 'value created')


@click.command('gc', help='Remove old images', cls=AsyncCommand)
@click.argument('count', type=click.INT, default=1)
@click.pass_obj
async def image_gc(env, count):
    if count < 0:
        click.echo('Count should be more or equal to 0')
        sys.exit(-1)
    known_repos = {i.repository for i in env.images}
    containers = await env.docker.containers(params={'all': 'true'})
    repo_tags_used = {c['Image'] for c in containers}

    by_repo = defaultdict(list)
    to_delete = []

    images = await env.docker.images()
    for image in images:
        repo_tags = set(image['RepoTags'])
        if repo_tags == {'<none>:<none>'}:
            to_delete.append(image['Id'])
            continue
        repo_tags.difference_update(repo_tags_used)
        for repo_tag in repo_tags:
            repo, _, tag = repo_tag.partition(':')
            if repo in known_repos:
                by_repo[repo].append(_Tag(tag, image['Created']))

    for repo, tags in by_repo.items():
        latest_tags = sorted(tags, key=attrgetter('created'), reverse=True)
        for tag in latest_tags[count:]:
            to_delete.append('{}:{}'.format(repo, tag.value))

    for image in to_delete:
        await env.docker.remove_image(image)
        click.echo('Removed: {}'.format(image))


async def _get_images_info(env):
    available = set()
    counts = Counter()
    sizes = {}
    images = await env.docker.images()
    for image in images:
        available.update(image['RepoTags'])
        for repo_tag in image['RepoTags']:
            repo, _ = repo_tag.split(':')
            counts[repo] += 1
            sizes[repo_tag] = image['VirtualSize']
    return available, counts, sizes


@click.command('list', help='List images', cls=AsyncCommand)
@click.pass_obj
async def image_list(env):
    from .._requires.tabulate import tabulate

    available, counts, sizes = await _get_images_info(env)
    images = sorted(env.images, key=lambda i: i.name)
    versions = image_versions(env.images, images)

    rows = []
    for image, version in zip(images, versions):
        di = DockerImage.from_image(image, version)
        if di.name in available:
            pretty_name = pretty('\u2714 {_green}{}{_r}', image.name)
        else:
            pretty_name = pretty('\u2717 {_red}{}{_r}', image.name)
        size = sizes.get(di.name, 0)
        pretty_size = format_size(size) if size else None
        count = counts.get(image.repository, None)
        rows.append([pretty_name, di.name, pretty_size, count])

    click.echo(tabulate(rows, headers=['  Image name', 'Docker image',
                                       'Size', 'Versions']))


def _image_ext_help(ctx, formatter):
    if ctx.obj.images:
        with formatter.section('Images'):
            formatter.write_dl([(image.name,
                                 image.description or '')
                                for image in ctx.obj.images])
    else:
        with formatter.section('Images'):
            formatter.write_text('--- not defined ---')


def create_images_cli():
    image_group = ExtGroup('image', help='Images creation and delivery',
                           ext_help=_image_ext_help)
    image_group.add_command(image_run)
    image_group.add_command(image_pull)
    image_group.add_command(image_push)
    image_group.add_command(image_build)
    image_group.add_command(image_list)
    image_group.add_command(image_gc)
    image_group.add_command(image_info)

    cli = click.Group()
    cli.add_command(image_group)
    return cli



import json
from contextlib import asynccontextmanager

from urllib.parse import urlencode

from .http import connect_unix
from .auth import read_config, server_name, resolve_auth, encode_header
from .utils import cached_property


CHUNK_SIZE = 65535


async def _recv_json(stream, response):
    content_type = response.headers.get(b'content-type')
    assert content_type == b'application/json', response

    if b'content-length' in response.headers:
        content_length = response.headers[b'content-length']
        data = await stream.recv_data(content_length=int(content_length))
    elif b'transfer-encoding' in response.headers:
        transfer_encoding = response.headers[b'transfer-encoding']
        assert transfer_encoding == b'chunked', response
        chunks = [c async for c in stream.recv_data_chunked()]
        data = b''.join(chunks)
    else:
        assert False, response

    return json.loads(data.decode('utf-8'))


async def _request_json(method, path, data=None, *, _ok_statuses=None):
    if _ok_statuses is None:
        _ok_statuses = frozenset({200, 201, 204})
    async with connect_unix() as stream:
        headers = [
            ('Host', 'localhost'),
            ('Connection', 'close'),
        ]
        if data is not None:
            json_data = json.dumps(data).encode('utf-8')
            headers.append(('Content-Type', 'application/json'))
            headers.append(('Content-Length', str(len(json_data))))
        await stream.send_request(method, path, headers,
                                  end_stream=(data is None))
        if data is not None:
            await stream.send_data(json_data)
        response = await stream.recv_response()
        if response.status_code == 204:
            return None
        if response.status_code in _ok_statuses:
            return await _recv_json(stream, response)
        else:
            raise response.error()


async def _get_json(path, *, _ok_statuses=None):
    return await _request_json('GET', path, _ok_statuses=_ok_statuses)


async def _post_json(path, data=None, *, _ok_statuses=None):
    return await _request_json('POST', path, data=data,
                               _ok_statuses=_ok_statuses)


async def _delete_json(path, *, _ok_statuses=None):
    return await _request_json('DELETE', path, _ok_statuses=_ok_statuses)


class Docker:

    @cached_property
    def _docker_config(self):
        return read_config()

    async def _auth_header(self, image_name):
        registry = server_name(image_name)
        auth = await resolve_auth(self._docker_config, registry)
        if auth is not None:
            auth_header = encode_header(auth)
            return auth_header
        else:
            return None

    async def images(self):
        return await _get_json('/images/json')

    async def create_container(self, spec, *, params=None):
        uri = '/containers/create'
        if params:
            uri += '?' + urlencode(params)
        return await _post_json(uri, spec)

    async def resize(self, id_, *, params=None):
        assert isinstance(id_, str), id_
        uri = '/containers/{id}/resize'.format(id=id_)
        if params:
            uri += '?' + urlencode(params)
        async with connect_unix() as stream:
            await stream.send_request('POST', uri, [
                ('Host', 'localhost'),
            ])
            response = await stream.recv_response()
            if response.status_code == 200:
                pass
            else:
                raise response.error()

    async def start(self, id_, *, params=None):
        assert isinstance(id_, str), id_
        uri = '/containers/{id}/start'.format(id=id_)
        if params:
            uri += '?' + urlencode(params)
        async with connect_unix() as stream:
            await stream.send_request('POST', uri, [
                ('Host', 'localhost'),
            ])
            response = await stream.recv_response()
            if response.status_code == 204:
                pass
            elif response.status_code == 304:
                pass
            else:
                raise response.error()

    async def exec_create(self, id_, spec):
        assert isinstance(id_, str), id_
        uri = '/containers/{id}/exec'.format(id=id_)
        return await _post_json(uri, spec)

    @asynccontextmanager
    async def exec_start(self, id_, spec, stdin_proto, stdout_proto):
        assert isinstance(id_, str), id_
        uri = '/exec/{id}/start'.format(id=id_)
        async with connect_unix(
            stdin_proto=stdin_proto, stdout_proto=stdout_proto
        ) as stream:
            json_data = json.dumps(spec).encode('utf-8')
            await stream.send_request('POST', uri, [
                ('Host', 'localhost'),
                ('Content-Type', 'application/json'),
                ('Content-Length', str(len(json_data))),
                ('Connection', 'Upgrade'),
                ('Upgrade', 'tcp'),
            ], end_stream=False)
            await stream.send_data(json_data)
            response = await stream.recv_response()
            if response.status_code == 101:
                yield stream.protocol
            else:
                raise response.error()

    async def exec_inspect(self, id_):
        assert isinstance(id_, str), id_
        uri = '/exec/{id}/json'.format(id=id_)
        return await _get_json(uri)

    @asynccontextmanager
    async def attach(self, id_, stdin_proto, stdout_proto, *, params=None):
        assert isinstance(id_, str), id_
        uri = '/containers/{id}/attach'.format(id=id_)
        if params:
            uri += '?' + urlencode(params)
        async with connect_unix(
            stdin_proto=stdin_proto, stdout_proto=stdout_proto
        ) as stream:
            await stream.send_request('POST', uri, [
                ('Host', 'localhost'),
                ('Connection', 'Upgrade'),
                ('Upgrade', 'tcp'),
            ])
            response = await stream.recv_response()
            if response.status_code == 101:
                yield stream.protocol
            else:
                raise response.error()

    async def remove_container(self, id_, *, params=None):
        assert isinstance(id_, str), id_
        uri = '/containers/{id}'.format(id=id_)
        if params:
            uri += '?' + urlencode(params)
        async with connect_unix() as stream:
            await stream.send_request('DELETE', uri, [
                ('Host', 'localhost'),
            ])
            response = await stream.recv_response()
            if response.status_code == 204:
                pass
            else:
                raise response.error()

    async def create_image(self, *, params=None):
        uri = '/images/create'
        if params:
            uri += '?' + urlencode(params)
        headers = [('Host', 'localhost')]
        if 'fromImage' in params:
            auth_header = await self._auth_header(params['fromImage'])
            if auth_header:
                headers.append(('X-Registry-Auth', auth_header))

        async with connect_unix() as stream:
            await stream.send_request('POST', uri, headers)
            response = await stream.recv_response()
            if response.status_code == 200:
                async for chunk in stream.recv_data_chunked():
                    yield chunk
            else:
                raise response.error()

    async def push(self, name, *, params):
        uri = '/images/{name}/push'.format(name=name)
        if params:
            uri += '?' + urlencode(params)
        headers = [('Host', 'localhost')]
        auth_header = await self._auth_header(name)
        if auth_header:
            headers.append(('X-Registry-Auth', auth_header))

        async with connect_unix() as stream:
            await stream.send_request('POST', uri, headers)
            response = await stream.recv_response()
            if response.status_code == 200:
                async for chunk in stream.recv_data_chunked():
                    yield chunk
            else:
                raise response.error()

    async def containers(self, *, params):
        uri = '/containers/json'
        if params:
            uri += '?' + urlencode(params)
        return await _get_json(uri)

    async def remove_image(self, name):
        uri = '/images/{name}'.format(name=name)
        return await _delete_json(uri)

    async def wait(self, id_):
        assert isinstance(id_, str), id_
        uri = '/containers/{id}/wait'.format(id=id_)
        return await _post_json(uri)

    async def stop(self, id_, *, params):
        assert isinstance(id_, str), id_
        uri = '/containers/{id}/stop'.format(id=id_)
        if params:
            uri += '?' + urlencode(params)
        await _post_json(uri)

    async def pause(self, id_):
        assert isinstance(id_, str), id_
        uri = '/containers/{id}/pause'.format(id=id_)
        await _post_json(uri)

    async def commit(self, *, params):
        uri = '/commit'
        if params:
            uri += '?' + urlencode(params)
        return await _post_json(uri)

    async def unpause(self, id_):
        assert isinstance(id_, str), id_
        uri = '/containers/{id}/unpause'.format(id=id_)
        await _post_json(uri)

    async def create_network(self, *, data):
        uri = '/networks/create'
        return await _post_json(uri, data=data)

    async def put_archive(self, id_, arch, *, params):
        uri = '/containers/{id}/archive'.format(id=id_)
        if params:
            uri += '?' + urlencode(params)
        headers = [
            ('Host', 'localhost'),
            ('transfer-encoding', 'chunked'),
        ]
        async with connect_unix() as stream:
            await stream.send_request('PUT', uri, headers, end_stream=False)
            while True:
                chunk = arch.read(CHUNK_SIZE)
                if len(chunk) == CHUNK_SIZE:
                    await stream.send_data(chunk, end_stream=False)
                else:
                    if chunk:
                        await stream.send_data(chunk)
                    else:
                        await stream.end()
                    break
            response = await stream.recv_response()
            if response.status_code != 200:
                raise response.error()



import sys
from collections import defaultdict


class Status:
    def __init__(self, *, output=sys.stdout):
        self._idx = []
        self._steps = defaultdict(list)
        self._titles = {}
        self._current = 0
        self._output = output

    def _up(self, count):
        self._output.write(f'{chr(27)}[{count}A')

    def _down(self, count):
        self._output.write(f'{chr(27)}[{count}B')

    def _erase(self):
        self._output.write(f'{chr(27)}[2K\r')

    def _move(self, to):
        count = to - self._current
        if count > 0:
            self._down(count)
        elif count < 0:
            self._up(-count)
        self._current = to

    def _newline(self):
        self._output.write('\n')
        self._current += 1

    def _refresh(self, key):
        self._move(self._idx.index(key))
        self._erase()
        self._output.write(self._titles[key])

    def add_task(self, title):
        self._move(len(self._idx))
        self._newline()

        key = object()
        self._idx.append(key)
        self._titles[key] = title
        self._refresh(key)

        self._output.flush()
        return key

    def add_step(self, task_key, title):
        if task_key in self._steps:
            append_after = self._steps[task_key][-1]
        else:
            append_after = task_key

        insert_pos = self._idx.index(append_after) + 1

        key = object()
        self._idx.insert(insert_pos, key)
        self._steps[task_key].append(key)
        self._titles[key] = title

        # rerender
        for key in self._idx[insert_pos:]:
            self._refresh(key)
        self._newline()

        self._output.flush()
        return key

    def update(self, key, title):
        self._titles[key] = title
        self._refresh(key)
        self._output.flush()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._move(len(self._idx))
        self._erase()
        self._output.flush()



import sys
import signal
import asyncio

from .._requires import click


SIGNALS = (signal.SIGINT, signal.SIGTERM)


class ExtGroup(click.Group):

    def __init__(self, *args, ext_help=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.ext_help = ext_help

    def parse_args(self, ctx, args):
        if not args:
            click.echo(ctx.get_help(), color=ctx.color)
            ctx.exit()
        return super().parse_args(ctx, args)

    def format_help_text(self, ctx, formatter):
        super().format_help_text(ctx, formatter)
        if self.ext_help is not None:
            self.ext_help(ctx, formatter)


def _exit_handler(sig_num):
    msg = 'Interrupted!' if sig_num == signal.SIGINT else 'Terminated!'
    click.echo(msg, file=sys.stderr, nl=True)
    raise SystemExit(128 + sig_num)


async def _invoke(callback, *args, **kwargs):
    loop = asyncio.get_running_loop()
    for sig_num in SIGNALS:
        loop.add_signal_handler(sig_num, _exit_handler, sig_num)
    try:
        return await callback(*args, **kwargs)
    finally:
        for sig_num in SIGNALS:
            loop.remove_signal_handler(sig_num)


def _async(callback):
    def wrapper(*args, **kwargs):
        return asyncio.run(_invoke(callback, *args, **kwargs))
    return wrapper


class AsyncCommand(click.Command):

    def __init__(self, *args, callback, **kwargs):
        super().__init__(*args, callback=_async(callback), **kwargs)


class AsyncProxyCommand(AsyncCommand):

    def parse_args(self, ctx, args):
        ctx.args = args

    def invoke(self, ctx):
        if self.callback is not None:
            ctx.invoke(self.callback, args=ctx.args)



import sys
import os.path
import logging
import asyncio

from ._requires import click

from .http import HTTPError


log = logging.getLogger(__name__)


class _VolumeBinds:

    def visit(self, obj):
        return obj.accept(self)

    def visit_RO(self, _):
        return 'ro'

    def visit_RW(self, _):
        return 'rw'

    def visit_localpath(self, obj):
        from_ = os.path.abspath(obj.from_)
        to = os.path.abspath(obj.to)
        # FIXME: implement proper errors reporting
        assert os.path.exists(from_),\
            'Local path does not exists: {}'.format(from_)
        return '{}:{}:{}'.format(from_, to, self.visit(obj.mode))

    def visit_namedvolume(self, obj):
        to = os.path.abspath(obj.to)
        return '{}:{}:{}'.format(obj.name, to, self.visit(obj.mode))


def _volumes(volumes):
    return {os.path.abspath(vol.to): {} for vol in volumes}


def _volume_binds(volumes):
    transformer = _VolumeBinds()
    return [transformer.visit(v) for v in volumes]


def _exposed_ports(ports):
    return {'{}/{}'.format(p.port, p.proto): {} for p in ports}


def _bind_to(port):
    return [{'HostPort': str(port.as_), 'HostIp': port.addr}]


def _port_binds(ports):
    return {
        '{}/{}'.format(port.port, port.proto): _bind_to(port)
        for port in ports
    }


async def start(docker, image, command, *, init=None, tty=True,
                entrypoint=None, volumes=None, ports=None, environ=None,
                work_dir=None, network=None, network_alias=None, label=None):
    spec = {
        'Image': image.name,
        'Cmd': command,
        'OpenStdin': True,
        'Tty': tty,
    }
    if ports:
        spec['ExposedPorts'] = _exposed_ports(ports)
    if environ:
        spec['Env'] = ['{}={}'.format(k, v) for k, v in (environ.items() or ())]
    if volumes:
        spec['Volumes'] = _volumes(volumes)
    if entrypoint is not None:
        spec['Entrypoint'] = entrypoint
    if work_dir:
        spec['WorkingDir'] = os.path.abspath(work_dir)
    if label:
        spec['Labels'] = {label: ''}

    host_config = {}
    if init:
        host_config['Init'] = True
    if volumes:
        host_config['Binds'] = _volume_binds(volumes)
    if ports:
        host_config['PortBindings'] = _port_binds(ports)
    if network:
        host_config['NetworkMode'] = network
    if host_config:
        spec['HostConfig'] = host_config

    networking_config = {}
    if network and network_alias:
        networking_config['EndpointsConfig'] = {
            network: {'Aliases': [network_alias]},
        }
    if networking_config:
        spec['NetworkingConfig'] = networking_config

    return await docker.create_container(spec)


async def start_service(docker, *args, **kwargs):
    c = await start(docker, *args, **kwargs)
    await docker.start(c['Id'])


async def resize(docker, id_):
    # TODO: maybe set also $LINES and $COLUMNS variables, add SIGWINCH handler
    width, height = click.get_terminal_size()
    try:
        await docker.resize(id_, params={'w': str(width), 'h': str(height)})
    except HTTPError as e:
        log.debug('Failed to resize terminal: %s', e)


class StdIOProtocol(asyncio.Protocol):
    transport: asyncio.Transport

    def __init__(self, http_proto=None):
        self.http_proto = http_proto

    def connection_made(self, transport):
        self.transport = transport

    def pause_writing(self):
        self.http_proto.transport.pause_reading()

    def resume_writing(self):
        self.http_proto.transport.resume_reading()

    def data_received(self, data):
        self.http_proto.transport.write(data)


async def attach(docker, id_):
    loop = asyncio.get_running_loop()
    stdin_proto = StdIOProtocol()
    await loop.connect_read_pipe(lambda: stdin_proto, sys.stdin)
    stdin_proto.transport.pause_reading()

    stdout_proto = StdIOProtocol()
    await loop.connect_write_pipe(lambda: stdout_proto, sys.stdout)

    async with docker.attach(
        id_, stdin_proto, stdout_proto,
        params={'logs': '1', 'stream': '1',
                'stdin': '1', 'stdout': '1', 'stderr': '1'}
    ) as http_proto:
        stdin_proto.http_proto = http_proto
        stdout_proto.http_proto = http_proto

        stdin_proto.transport.resume_reading()
        await resize(docker, id_)
        await http_proto.wait_closed()


async def run(docker, tty, image, command, *, init=None,
              volumes=None, ports=None, environ=None, work_dir=None,
              network=None, network_alias=None):
    c = await start(docker, image, command, init=init, tty=tty,
                    volumes=volumes,
                    ports=ports, environ=environ, work_dir=work_dir,
                    network=network, network_alias=network_alias)
    try:
        await docker.start(c['Id'])
        await attach(docker, c['Id'])
        exit_code = await docker.wait(c['Id'])
        return exit_code['StatusCode']
    finally:
        await docker.remove_container(c['Id'],
                                      params={'v': 'true', 'force': 'true'})



from .utils import cached_property, SequenceMap
from .docker import Docker


class Environ:
    docker: Docker

    def __init__(self, meta, images, services):
        self._meta = meta
        self.images = SequenceMap(images, lambda i: i.name)
        self.services = SequenceMap(services, lambda i: i.name)

    @property
    def namespace(self):
        return self._meta.namespace or 'default'

    @property
    def network(self):
        return 'pi-{}'.format(self.namespace)

    @cached_property
    def docker(self):
        return Docker()



import re
import json
import base64
import codecs
import os.path
import asyncio
import subprocess


_PREFIX = 'docker-credential-'


def read_config():
    path = os.path.expanduser('~/.docker/config.json')
    with codecs.open(path, encoding='utf-8') as f:
        json_data = f.read()
    return json.loads(json_data)


async def _read_creds(creds_store, server):
    if not re.match(r'^\w+$', creds_store, re.ASCII):
        raise ValueError('Invalid credsStore: {!r}'.format(creds_store))

    proc = await asyncio.create_subprocess_exec(
        _PREFIX + creds_store, 'get',
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    stdout, stderr = await proc.communicate(server.encode('ascii'))
    if proc.returncode != 0:
        return None
    else:
        data = json.loads(stdout)
        return {
            'Username': data['Username'],
            'Password': data['Secret'],
            'ServerAddress': server,
        }


def _decode_auth(auth_data, server):
    auth_data_decoded = base64.b64decode(auth_data)
    username, _, password = auth_data_decoded.partition(':')
    return {
        'Username': username,
        'Password': password,
        'ServerAddress': server,
    }


async def resolve_auth(config, server):
    config_auths = config.get('auths')
    if config_auths is None:
        return None

    server_auth = config_auths.get(server)
    if server_auth is not None:
        auth_data = server_auth.get('auth')
        if auth_data is not None:
            return _decode_auth(auth_data, server)

    creds_store = config.get('credsStore')
    if creds_store is not None:
        return await _read_creds(creds_store, server)

    return None


def server_name(image_name):
    registry, _, name = image_name.partition('/')
    if not name:
        return 'docker.io'
    else:
        return registry


def encode_header(auth):
    json_data = json.dumps(auth)
    return base64.urlsafe_b64encode(json_data.encode('ascii'))



import json
import hashlib

from .http import HTTPError
from .types import DockerImage, Image, ActionType


class Hasher:

    def visit(self, obj):
        return obj.accept(self)

    def visit_image(self, obj):
        yield obj.repository.encode('utf-8')
        for task in obj.tasks:
            yield from self.visit(task)

    def visit_task(self, obj):
        yield obj.run.encode('utf-8')
        for value in obj.where.values():
            if isinstance(value, ActionType):
                yield from self.visit(value)
            else:
                yield str(value).encode('utf-8')

    def visit_download(self, obj):
        yield obj.url.encode('utf-8')

    def visit_file(self, obj):
        with open(obj.path, 'rb') as f:
            while True:
                chunk = f.read(2**16)
                if not chunk:
                    break
                else:
                    yield chunk

    def visit_bundle(self, obj):
        yield obj.path.encode('utf-8')


def image_hashes(images_map, images, *, _cache=None):
    _cache = _cache or {}

    hasher = Hasher()
    hashes = []
    for image in images:
        if image.name in _cache:
            hashes.append(_cache[image.name])
            continue

        if isinstance(image.from_, DockerImage):
            parent_hashable = image.from_.name
        else:
            parent = images_map.get(image.from_)
            parent_hashable, = image_hashes(images_map, [parent],
                                            _cache=_cache)

        h = hashlib.sha1()
        h.update(parent_hashable.encode('utf-8'))
        for chunk in hasher.visit(image):
            h.update(chunk)
        hex_digest = _cache[image.name] = h.hexdigest()
        hashes.append(hex_digest)
    return hashes


def image_versions(images_map, images):
    hashes = image_hashes(images_map, images)
    return [h[:12] for h in hashes]


def docker_image(images_map, image):
    if isinstance(image, str):
        image = images_map.get(image)
        version, = image_versions(images_map, [image])
        return DockerImage.from_image(image, version)
    elif isinstance(image, DockerImage):
        return image
    else:
        raise TypeError(repr(image))


def resolve_deps(deps):
    while deps:
        resolved = set()
        for name, parent_name in deps.items():
            if parent_name not in deps:
                resolved.add(name)
        if not resolved:
            raise TypeError('Images hierarchy build error, '
                            'circular dependency found in these images: {}'
                            .format(', '.join(sorted(deps.keys()))))
        for name in resolved:
            yield name, deps[name]
        deps = {k: v for k, v in deps.items() if k not in resolved}


def get_images(config):
    return [i for i in config if isinstance(i, Image)]


def _process_pull_progress(status, image):
    key = status.add_task('=> Pulling image {}'.format(image))
    steps = {}
    while True:
        event = yield
        if event.get('status', '').startswith('Pulling from '):
            continue
        if 'id' in event:
            title = '  [{}] '.format(event['id']) + event['status']
            if 'progress' in event:
                title += ': ' + event['progress']
            if event['id'] in steps:
                status.update(steps[event['id']], title)
            else:
                steps[event['id']] = status.add_step(key, title)


def _process_push_progress(status, image):
    key = status.add_task('=> Pushing image {}'.format(image))
    steps = {}
    while True:
        event = yield
        if 'id' in event:
            title = '  [{}] '.format(event['id']) + event['status']
            if 'progress' in event:
                title += ': ' + event['progress']
            if event['id'] in steps:
                status.update(steps[event['id']], title)
            else:
                steps[event['id']] = status.add_step(key, title)


async def pull(docker, docker_image_: DockerImage, *, status):
    repository, _, tag = docker_image_.name.partition(':')
    params = {'fromImage': repository, 'tag': tag}
    try:
        gen = _process_pull_progress(status, docker_image_.name)
        gen.send(None)
        async for chunk in docker.create_image(params=params):
            for doc in chunk.decode('utf-8').splitlines():
                gen.send(json.loads(doc))
    except HTTPError:
        return False
    else:
        return True


async def push(docker, docker_image_, *, status):
    name, _, tag = docker_image_.name.partition(':')
    params = {'tag': tag}
    try:
        gen = _process_push_progress(status, docker_image_.name)
        gen.send(None)
        async for chunk in docker.push(name, params=params):
            for doc in chunk.decode('utf-8').splitlines():
                gen.send(json.loads(doc))
    except HTTPError:
        return False
    else:
        return True



import io
import os
import sys
import math
import uuid
import tarfile
import logging
import hashlib
import asyncio
import tempfile
import unicodedata
from typing import Optional
from pathlib import Path
from asyncio import wait, Queue, Event, gather, FIRST_EXCEPTION, WriteTransport
from dataclasses import dataclass, field
from urllib.parse import urlsplit
from concurrent.futures import ProcessPoolExecutor

from ._requires import jinja2

from .run import StdIOProtocol
from .http import connect_tcp
from .types import ActionType
from .utils import terminate
from .images import docker_image, image_versions


log = logging.getLogger(__name__)


@dataclass
class Result:
    file: tempfile.NamedTemporaryFile
    uuid: str = field(default_factory=lambda: uuid.uuid4().hex)

    def value(self):
        return hashlib.sha1(self.file.name.encode('utf-8')).hexdigest()

    def close(self):
        self.file.close()


@dataclass
class ActionState:
    complete: Event
    result: Result
    error: Optional[str] = None


async def _download(url, output, *, _max_redirects=5):
    redirects = 0
    initial_url = url
    initial_secure = None
    while redirects < _max_redirects:
        url_parts = urlsplit(url)

        secure = True if url_parts.scheme == 'https' else False
        if redirects == 0:
            initial_secure = secure
        if initial_secure is True and secure is False:
            raise Exception(f'Redirect to insecure url: {url}')

        host, _, port = url_parts.netloc.partition(':')
        if not port:
            port = 443 if url_parts.scheme == 'https' else 80
        else:
            port = int(port)
        path = url_parts.path
        if url_parts.query:
            path += '?' + url_parts.query

        async with connect_tcp(host, port, secure=secure) as stream:
            await stream.send_request('GET', path, [
                ('host', url_parts.netloc),
            ])
            response = await stream.recv_response()
            if response.status_code in {301, 302}:
                url = response.headers[b'location'].decode('ascii')
                redirects += 1
                continue
            elif response.status_code != 200:
                response.error()
            async for chunk in stream.recv_data_chunked():
                output.write(chunk)
            break
    else:
        raise Exception(f'More than {_max_redirects} redirects: {initial_url}')


async def download(url, file_name, destination):
    with tempfile.NamedTemporaryFile() as tmp:
        with tarfile.open(file_name, mode='w:tar') as tar:
            await _download(url, tmp)
            tmp.seek(0)
            tar.addfile(tar.gettarinfo(arcname=destination, fileobj=tmp),
                        fileobj=tmp)


def file_(path, file_name, destination):
    file_path = Path(path).resolve()
    assert file_path.is_file(), file_path

    cur_path = Path('.').resolve()
    assert cur_path in file_path.parents, file_path

    with open(str(file_path), 'rb') as f:
        with tarfile.open(file_name, mode='w:tar') as tar:
            tar.addfile(tar.gettarinfo(arcname=destination, fileobj=f),
                        fileobj=f)


def bundle(action_path, file_name, destination):
    dir_path = Path(action_path).resolve()  # FIXME: raises FileNotFoundError
    assert dir_path.is_dir(), dir_path

    cur_path = Path('.').resolve()
    assert cur_path in dir_path.parents, dir_path

    def _arc_path(path_):
        return os.path.join(destination, unicodedata.normalize('NFC', path_))

    with tarfile.open(file_name, mode='w:tar') as tar:
        for abs_path, _, names in os.walk(str(dir_path)):
            rel_path = Path(abs_path).relative_to(dir_path)
            if rel_path != Path('.'):
                tar.addfile(tar.gettarinfo(abs_path, _arc_path(str(rel_path))))

            for name in names:
                file_abs_path = os.path.join(abs_path, name)
                file_rel_path = str(rel_path.joinpath(name))
                with open(file_abs_path, 'rb') as f:
                    tar.addfile(
                        tar.gettarinfo(arcname=_arc_path(str(file_rel_path)),
                                       fileobj=f),
                        fileobj=f,
                    )


def iter_actions(task):
    for value in task.where.values():
        if isinstance(value, ActionType):
            yield value


def get_action_states(tasks):
    actions = {}
    for task in tasks:
        for action in iter_actions(task):
            if action in actions:
                continue
            result = Result(tempfile.NamedTemporaryFile())
            actions[action] = ActionState(Event(), result)
    return actions


async def wait_actions(states):
    if states:
        await gather(*[state.complete.wait() for state in states.values()])


class ActionDispatcher:

    def __init__(self, states, io_queue, cpu_queue):
        self.states = states
        self.io_queue = io_queue
        self.cpu_queue = cpu_queue

    @classmethod
    async def dispatch(cls, states, io_queue, cpu_queue):
        dispatcher = cls(states, io_queue, cpu_queue)
        for action in states:
            await dispatcher.visit(action)

    def visit(self, obj):
        return obj.accept(self)

    async def visit_download(self, obj):
        await self.io_queue.put((obj, self.states[obj]))

    async def visit_file(self, obj):
        await self.cpu_queue.put((obj, self.states[obj]))

    async def visit_bundle(self, obj):
        await self.cpu_queue.put((obj, self.states[obj]))


class IOExecutor:

    def visit(self, action):
        return action.accept(self)

    async def download(self, action, state):
        try:
            await download(
                action.url,
                state.result.file.name,
                state.result.uuid,
            )
        except Exception as err:
            log.debug('Download action failed: %r', action, exc_info=True)
            state.error = str(err)
            raise
        finally:
            state.complete.set()

    def visit_download(self, obj):
        return self.download


class CPUExecutor:

    def __init__(self, process_pool):
        self.process_pool = process_pool

    def visit(self, action):
        return action.accept(self)

    async def file(self, action, state):
        try:
            await asyncio.get_running_loop().run_in_executor(
                self.process_pool,
                file_, action.path, state.result.file.name, state.result.uuid
            )
        except Exception as err:
            log.debug('File action failed: %r', action, exc_info=True)
            state.error = str(err)
        finally:
            state.complete.set()

    async def bundle(self, action, state):
        try:
            await asyncio.get_running_loop().run_in_executor(
                self.process_pool,
                bundle, action.path, state.result.file.name, state.result.uuid
            )
        except Exception as err:
            log.debug('Bundle action failed: %r', action, exc_info=True)
            state.error = str(err)
        finally:
            state.complete.set()

    def visit_file(self, obj):
        return self.file

    def visit_bundle(self, obj):
        return self.bundle


async def worker(queue, executor):
    while True:
        action, state = await queue.get()
        process = executor.visit(action)
        await process(action, state)


async def pool(queue, executor, concurrency=2):
    loop = asyncio.get_running_loop()
    pending = [loop.create_task(worker(queue, executor))
               for _ in range(concurrency)]
    try:
        # fast exit on first error
        _, pending = await wait(pending, return_when=FIRST_EXCEPTION)
    finally:
        for task in pending:
            await terminate(task)


def task_cmd(task, results):
    ctx = {key: value if not isinstance(value, ActionType) else results[value]
           for key, value in task.where.items()}
    t = jinja2.Template(task.run)
    return t.render(ctx)


class WriteBuffer(WriteTransport):

    def __init__(self):
        super().__init__()
        self._buffer = io.BytesIO()

    def write(self, data):
        self._buffer.write(data)

    def dump(self):
        return self._buffer.getvalue()


async def _exec(docker, id_, cmd):
    if isinstance(cmd, str):
        cmd = ['/bin/sh', '-c', cmd]

    stdout_buffer = WriteBuffer()
    stdout_proto = StdIOProtocol()
    stdout_proto.connection_made(stdout_buffer)

    exec_ = await docker.exec_create(id_, {
        'Cmd': cmd,
        'AttachStdout': True,
        'AttachStderr': True,
    })
    async with docker.exec_start(
        exec_['Id'], {}, None, stdout_proto
    ) as http_proto:
        await http_proto.wait_closed()
    info = await docker.exec_inspect(exec_['Id'])
    exit_code = info['ExitCode']
    if exit_code:
        # FIXME: proper output
        print(stdout_buffer.dump().decode('utf-8', 'backslashreplace'),
              file=sys.stderr)
    return exit_code


async def build_image(docker, images_map, image, *, status):
    loop = asyncio.get_running_loop()
    version, = image_versions(images_map, [image])
    from_ = docker_image(images_map, image.from_)

    task_key = status.add_task('=> Building image {}:{} ({})'
                               .format(image.repository, version, image.name))

    io_queue = Queue()
    io_executor = IOExecutor()

    process_pool = ProcessPoolExecutor()
    cpu_queue = Queue()
    cpu_executor = CPUExecutor(process_pool)

    states = get_action_states(image.tasks)
    submitted_states = set()

    io_pool_task = loop.create_task(pool(io_queue, io_executor))
    cpu_pool_task = loop.create_task(pool(cpu_queue, cpu_executor))

    c = await docker.create_container({
        'Image': from_.name,
        'Cmd': '/bin/sh',
        'Tty': True,
        'AttachStdout': False,
        'AttachStderr': False,
    })
    try:
        await docker.start(c['Id'])
        exit_code = await _exec(docker, c['Id'], ['mkdir', '/.pi'])
        if exit_code:
            return False

        await ActionDispatcher.dispatch(states, io_queue, cpu_queue)

        total = len(image.tasks)
        padding = math.ceil(math.log10(total + 1))

        for i, task in enumerate(image.tasks, 1):
            task_states = {action: states[action]
                           for action in iter_actions(task)}

            await wait_actions(task_states)

            errors = {action: state.error
                      for action, state in task_states.items()
                      if state.error is not None}
            if errors:
                raise Exception(repr(errors))

            task_results = {action: '/.pi/{}'.format(state.result.uuid)
                            for action, state in task_states.items()}

            for action, state in task_states.items():
                if action not in submitted_states:
                    with open(state.result.file.name, 'rb') as tar:
                        await docker.put_archive(c['Id'], tar, params={
                            'path': '/.pi',
                        })
                    submitted_states.add(action)

            cmd = task_cmd(task, task_results)
            current_index = '{{:{}d}}'.format(padding).format(i)
            status.add_step(
                task_key, '[{}/{}] {}'.format(current_index, total, cmd),
            )
            exit_code = await _exec(docker, c['Id'], cmd)
            if exit_code:
                return False

        exit_code = await _exec(docker, c['Id'], ['rm', '-rf', '/.pi'])
        if exit_code:
            return False

        await docker.pause(c['Id'])
        await docker.commit(params={
            'container': c['Id'],
            'repo': image.repository,
            'tag': version,
        })
        await docker.unpause(c['Id'])
        return True

    finally:
        await terminate(io_pool_task)
        await terminate(cpu_pool_task)
        process_pool.shutdown()
        for state in states.values():
            state.result.close()
        await docker.remove_container(c['Id'],
                                      params={'v': 'true', 'force': 'true'})



import sys

from .._requires import click
from .._requires.tabulate import tabulate

from ..run import start_service
from ..utils import search_container, sh_to_list
from ..images import docker_image
from ..network import ensure_network
from ..console import pretty
from ..services import get_volumes, service_label

from .common import ExtGroup, AsyncCommand


@click.command('start', help='Start service', cls=AsyncCommand)
@click.argument('name')
@click.pass_obj
async def service_start(env, name):
    service = env.services.get(name, None)
    if service is None:
        click.echo('Unknown service name: {}'.format(name))
        sys.exit(-1)

    label = service_label(env.namespace, service)
    containers = await env.docker.containers(params={'all': 'true'})
    container = next(search_container(label, containers), None)
    if container is not None:
        if container['State'] == 'running':
            click.echo('Service is already running')
        elif container['State'] == 'exited':
            await env.docker.start(container['Id'])
            click.echo('Started previously stopped service')
        else:
            raise NotImplementedError(container['State'])
    else:
        exec_ = sh_to_list(service.exec) if service.exec else None
        args = sh_to_list(service.args) if service.args else None
        di = docker_image(env, service.image)
        await ensure_network(env.docker, env.network)
        await start_service(
            env.docker, di, args,
            entrypoint=exec_,
            volumes=get_volumes(service.volumes),
            ports=service.ports,
            environ=service.environ,
            network=env.network,
            network_alias=service.network_name or service.name,
            label=label,
        )
        click.echo('Service started')


@click.command('stop', help='Stop service', cls=AsyncCommand)
@click.argument('name')
@click.pass_obj
async def service_stop(env, name):
    service = env.services.get(name, None)
    if service is None:
        click.echo('Unknown service name: {}'.format(name))
        sys.exit(-1)

    label = service_label(env.namespace, service)
    all_containers = await env.docker.containers(params={'all': 'true'})
    containers = list(search_container(label, all_containers))
    if not containers:
        click.echo('Service was not started')
        sys.exit(-1)

    for container in containers:
        if container['State'] == 'running':
            await env.docker.stop(container['Id'], params={'t': '3'})
        await env.docker.remove_container(container['Id'],
                                          params={'v': 'true', 'force': 'true'})
    click.echo('Service stopped')


@click.command('status', help='Display services status', cls=AsyncCommand)
@click.pass_obj
async def service_status(env):
    containers = await env.docker.containers(params={'all': 'true'})

    running = set()
    exited = set()
    images = {}
    for container in containers:
        if container['State'] == 'exited':
            exited.update(container['Labels'])
        elif container['State'] == 'running':
            running.update(container['Labels'])
        for label in container['Labels']:
            images[label] = container['Image']

    rows = []
    for service in env.services:
        label = service_label(env.namespace, service)
        if label in running:
            status = pretty('{_green}running{_r}')
            image = images[label]
        elif label in exited:
            status = pretty('{_red}stopped{_r}')
            image = images[label]
        else:
            status = None
            image = None

        di = docker_image(env.images, service.image)

        if image is not None and image != di.name:
            image += ' (obsolete)'

        rows.append([service.name, status, image])
    click.echo(tabulate(rows, headers=['Service name', 'Status',
                                       'Docker image']))


def _service_ext_help(ctx, formatter):
    if ctx.obj.services:
        with formatter.section('Services'):
            formatter.write_dl([(service.name, service.description or '')
                                for service in ctx.obj.services])
    else:
        with formatter.section('Services'):
            formatter.write_text('--- not defined ---')


def create_service_cli():
    service_group = ExtGroup('service', help='Services status and management',
                             ext_help=_service_ext_help)
    service_group.add_command(service_start)
    service_group.add_command(service_stop)
    service_group.add_command(service_status)

    cli = click.Group()
    cli.add_command(service_group)
    return cli



from enum import Enum
from typing import Optional, Union, Any, Sequence, Mapping
from dataclasses import dataclass, field, MISSING
from collections import OrderedDict

from .utils import ImmutableDict, cached_property


class EnumConstruct:

    @classmethod
    def construct(cls, loader, node):
        assert not loader.construct_scalar(node),\
            '{}: No arguments expected'.format(cls.__name__)
        return cls(node.tag)


class ScalarConstruct:

    @classmethod
    def construct(cls, loader, node):
        return cls(loader.construct_scalar(node))


class SequenceConstruct:

    @classmethod
    def construct(cls, loader, node):
        return cls(loader.construct_sequence(node, deep=True))


class MappingConstruct:
    __rename_to__ = ImmutableDict()

    @classmethod
    def construct(cls, loader, node):
        rename_from = {v: k for k, v in cls.__rename_to__.items()}

        params = loader.construct_mapping(node, deep=True)

        required = {
            rename_from.get(f.name, f.name)
            for f in cls.__dataclass_fields__.values()
            if f.default is MISSING and f.default_factory is MISSING
        }
        optional = {
            rename_from.get(f.name, f.name)
            for f in cls.__dataclass_fields__.values()
            if f.default is not MISSING or f.default_factory is not MISSING
        }

        missing = set(required).difference(params)
        if missing:
            raise TypeError('Missing params {!r} for {!r}'.format(missing, cls))

        clean_params = {cls.__rename_to__.get(k, k): v
                        for k, v in params.items()
                        if k in required or k in optional}

        return cls(**clean_params)


@dataclass(frozen=True)
class Meta(MappingConstruct):
    __tag__ = '!Meta'

    namespace: Optional[str] = None
    description: Optional[str] = None

    def accept(self, visitor):
        return visitor.visit_meta(self)


@dataclass(frozen=True)
class DockerImage(ScalarConstruct):
    __tag__ = '!DockerImage'

    name: str

    def accept(self, visitor):
        return visitor.visit_dockerimage(self)

    @classmethod
    def from_image(cls, image, version):
        return cls('{}:{}'.format(image.repository, version))


@dataclass(frozen=True)
class Task:
    run: str
    where: Mapping[str, Any]

    def accept(self, visitor):
        return visitor.visit_task(self)

    @classmethod
    def from_config(cls, d):
        run = d['run']
        where = OrderedDict((k, v) for k, v in d.items()
                            if k != 'run')
        return cls(run, where)


@dataclass(frozen=True)
class Image(MappingConstruct):
    __tag__ = '!Image'
    __rename_to__ = ImmutableDict([
        ('from', 'from_'),
        ('tasks', '_tasks'),
    ])

    name: str
    repository: str
    from_: Optional[Union[str, DockerImage]] = None
    _tasks: Sequence[Any] = field(default=(), hash=False)
    description: Optional[str] = None

    def accept(self, visitor):
        return visitor.visit_image(self)

    @cached_property
    def tasks(self) -> Sequence[Task]:
        return tuple(Task.from_config(d) for d in self._tasks)


class Mode(EnumConstruct, Enum):
    RO = '!RO'
    RW = '!RW'

    def accept(self, visitor):
        return getattr(visitor, 'visit_{}'.format(self.name))(self)


class VolumeType:

    def accept(self, visitor):
        raise NotImplementedError


@dataclass(frozen=True)
class LocalPath(VolumeType, MappingConstruct):
    __tag__ = '!LocalPath'
    __rename_to__ = ImmutableDict([
        ('from', 'from_'),
    ])

    from_: str
    to: str
    mode: Mode = Mode.RO

    def accept(self, visitor):
        return visitor.visit_localpath(self)


@dataclass(frozen=True)
class NamedVolume(VolumeType, MappingConstruct):
    __tag__ = '!NamedVolume'

    name: str
    to: str
    mode: Mode = Mode.RO

    def accept(self, visitor):
        return visitor.visit_namedvolume(self)


@dataclass(frozen=True)
class Expose(MappingConstruct):
    __tag__ = '!Expose'
    __rename_to__ = ImmutableDict([
        ('as', 'as_'),
    ])

    port: int
    as_: int
    addr: str = '127.0.0.1'
    proto: str = 'tcp'  # TODO: replace with enum

    def accept(self, visitor):
        return visitor.visit_expose(self)


@dataclass(frozen=True)
class Service(MappingConstruct):
    __tag__ = '!Service'
    __rename_to__ = ImmutableDict([
        ('network-name', 'network_name'),
    ])

    name: str
    image: Union[str, DockerImage]
    exec: Optional[Union[str, Sequence[str]]] = None
    args: Optional[Union[str, Sequence[str]]] = None
    volumes: Optional[Sequence[VolumeType]] = None
    ports: Optional[Sequence[Expose]] = None
    environ: Optional[Mapping[str, str]] = None
    requires: Optional[Sequence[str]] = None
    network_name: Optional[str] = None
    description: Optional[str] = None

    def accept(self, visitor):
        return visitor.visit_service(self)


@dataclass(frozen=True)
class ParameterType:
    name: str
    type: Optional[str] = None
    default: Any = None

    def accept(self, visitor):
        raise NotImplementedError


class Argument(ParameterType, MappingConstruct):
    __tag__ = '!Argument'

    def accept(self, visitor):
        return visitor.visit_argument(self)


class Option(ParameterType, MappingConstruct):
    __tag__ = '!Option'

    def accept(self, visitor):
        return visitor.visit_option(self)


@dataclass(frozen=True)
class Command(MappingConstruct):
    __tag__ = '!Command'
    __rename_to__ = ImmutableDict([
        ('network-name', 'network_name'),
    ])

    name: str
    image: Union[str, DockerImage]
    run: Union[str, Sequence[str]]
    params: Optional[Sequence[ParameterType]] = None
    volumes: Optional[Sequence[VolumeType]] = None
    ports: Optional[Sequence[Expose]] = None
    environ: Optional[Mapping[str, str]] = None
    requires: Optional[Sequence[str]] = None
    network_name: Optional[str] = None
    description: Optional[str] = None

    def accept(self, visitor):
        return visitor.visit_command(self)


class ActionType:

    def accept(self, visitor):
        raise NotImplementedError


@dataclass(frozen=True)
class Download(ActionType, ScalarConstruct):
    __tag__ = '!Download'

    url: str

    def accept(self, visitor):
        return visitor.visit_download(self)


@dataclass(frozen=True)
class File(ActionType, ScalarConstruct):
    __tag__ = '!File'

    path: str

    def accept(self, visitor):
        return visitor.visit_file(self)


@dataclass(frozen=True)
class Bundle(ActionType, ScalarConstruct):
    __tag__ = '!Bundle'

    path: str

    def accept(self, visitor):
        return visitor.visit_bundle(self)



import ssl
import socket
import asyncio
from asyncio import Event

from typing import cast, NamedTuple
from contextlib import asynccontextmanager

from .utils import Wrapper
from ._requires import h11


if hasattr(socket, 'TCP_NODELAY'):
    _sock_type_mask = 0xf if hasattr(socket, 'SOCK_NONBLOCK') else 0xffffffff

    def _set_nodelay(sock):
        if (
            sock.family in {socket.AF_INET, socket.AF_INET6}
            and sock.type & _sock_type_mask == socket.SOCK_STREAM
            and sock.proto == socket.IPPROTO_TCP
        ):
            sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
else:
    def _set_nodelay(sock):
        pass


class HTTPError(Exception):

    def __init__(self, reason):
        super().__init__(reason)
        self.reason = reason


class Response(NamedTuple):
    status_code: int
    headers: dict
    reason: bytes

    def error(self):
        try:
            reason = self.reason.decode('ascii')
        except UnicodeDecodeError:
            reason = repr(self.reason)
        raise HTTPError(reason)


class Stream:

    def __init__(self, protocol, connection: h11.Connection,
                 transport: asyncio.Transport):
        self.protocol = protocol  # FIXME: circular reference
        self.connection = connection
        self.transport = transport

        self._response = None
        self._response_waiter = asyncio.Event()
        self._data = []
        self._data_size = 0
        self._data_waiter = asyncio.Event()
        self._eof = False

        self._wrapper = Wrapper()

    async def send_request(self, method, path, headers, *, end_stream=True):
        data = self.connection.send(h11.Request(method=method, target=path,
                                                headers=headers))
        self.transport.write(data)
        if end_stream:
            await self.end()

    async def send_data(self, data, *, end_stream=True):
        data = self.connection.send(h11.Data(data=data))
        self.transport.write(data)
        if end_stream:
            await self.end()

    async def recv_response(self):
        with self._wrapper:
            await self._response_waiter.wait()
            return Response(self._response.status_code,
                            dict(self._response.headers),
                            self._response.reason)

    async def recv_data(self, content_length):
        with self._wrapper:
            while True:
                await self._data_waiter.wait()
                if self._data_size < content_length:
                    self._data_waiter.clear()
                else:
                    assert self._eof
                    assert self._data_size == content_length
                    return b''.join(self._data)

    async def recv_data_chunked(self):
        with self._wrapper:
            while True:
                await self._data_waiter.wait()
                for chunk in self._data:
                    yield chunk
                del self._data[:]
                if self._eof:
                    break
                else:
                    self._data_waiter.clear()

    async def end(self):
        data = self.connection.send(h11.EndOfMessage())
        self.transport.write(data)

    def __response__(self, response: h11.Response):
        self._response = response
        self._response_waiter.set()

    def __data__(self, data: h11.Data):
        self._data.append(data.data)
        self._data_size += len(data.data)
        self._data_waiter.set()

    def __end__(self):
        self._eof = True
        self._data_waiter.set()

    def __terminated__(self):
        if not self._eof:
            self._wrapper.cancel(Exception('Connection closed'))


class HTTPProtocol(asyncio.Protocol):
    connection = None
    transport: asyncio.Transport = None
    stream: Stream = None

    hijacked = False

    def __init__(self, *, stdin_proto=None, stdout_proto=None):
        self._stdin_proto = stdin_proto
        self._stdout_proto = stdout_proto
        self._closed = Event()

    def connection_made(self, transport):
        sock = transport.get_extra_info('socket')
        if sock is not None:
            _set_nodelay(sock)

        self.connection = h11.Connection(h11.CLIENT)
        self.transport = transport
        self.stream = Stream(self, self.connection, self.transport)

    def pause_writing(self):
        if self.hijacked:
            assert self._stdin_proto
            self._stdin_proto.transport.pause_reading()

    def resume_writing(self):
        if self.hijacked:
            assert self._stdin_proto
            self._stdin_proto.transport.resume_reading()

    def data_received(self, data: bytes):
        if self.hijacked:
            assert self._stdout_proto
            self._stdout_proto.transport.write(data)
        else:
            self.connection.receive_data(data)
            while True:
                event = self.connection.next_event()
                # print(event)
                event_type = type(event)
                # print(event_type)
                if event_type is h11.Response:
                    self.stream.__response__(event)
                elif event_type is h11.InformationalResponse:
                    self.stream.__response__(event)
                    if event.status_code == 101:
                        self.hijacked = True
                elif event_type is h11.Data:
                    self.stream.__data__(event)
                elif event_type is h11.EndOfMessage:
                    self.stream.__end__()
                elif event is h11.NEED_DATA:
                    break
                elif event is h11.PAUSED:
                    break

    def connection_lost(self, exc):
        if not self.hijacked:
            self.stream.__terminated__()
        self.transport.close()
        self._closed.set()

    async def wait_closed(self):
        return await self._closed.wait()


@asynccontextmanager
async def connect_unix(*, stdin_proto=None, stdout_proto=None):
    loop = asyncio.get_running_loop()
    transport, protocol = await loop.create_unix_connection(
        lambda: HTTPProtocol(stdin_proto=stdin_proto,
                             stdout_proto=stdout_proto),
        '/var/run/docker.sock',
    )
    try:
        yield cast(HTTPProtocol, protocol).stream
    finally:
        transport.close()


@asynccontextmanager
async def connect_tcp(host, port, *, secure=False):
    loop = asyncio.get_running_loop()
    ssl_context = ssl.create_default_context() if secure else None
    transport, protocol = await loop.create_connection(
        HTTPProtocol, host, port, ssl=ssl_context,
    )
    try:
        yield cast(HTTPProtocol, protocol).stream
    finally:
        transport.close()
